/*
 * srtp+gdoi.c
 *
 * GDOI key management interface for SRTP
 *
 * David A. McGrew and Brian E. Weis
 * Cisco Systems, Inc.
 */
/*
 *	
 * Copyright  (c) 2001, 2002, Cisco Systems, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.
 * 
 *   Neither the name of the Cisco Systems, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "rtp.h"

unsigned char *
srtp_gdoi_get_policy_and_keys(unsigned int, 
			      struct sockaddr_in *, unsigned long,
			      gdoi_policy_t *, gdoi_keys_t *);

int
srtp_gdoi_crypto_init (srtp_ctx_t *ctx, rtp_msg_t *message,
		       sec_serv_t sec_servs,
		       unsigned char *input_key, gdoi_policy_t *policy, 
		       gdoi_keys_t *keys) {
  unsigned char key[MAX_KEY_LEN];
  unsigned char salt[MAX_KEY_LEN];
  unsigned short computed_salt_len;
  err_status_t stat;

  if (input_key) {
    /* use manual keying */
    if ((is_hex_string(input_key) != 1) || (strlen(input_key) != 64)) {
      fprintf(stderr, "key should be 64 hexadecimal digits\n");
      return 1;
    }
    hex_string_to_octet_string(key, input_key, 16);
    hex_string_to_octet_string(salt, input_key + 32, 16);
    printf("set encryption key/salt to %s/", octet_string_hex_string(key, 32));
    printf("%s\n", octet_string_hex_string(salt, 32));
    
  } else if (!keys || !policy)  {
    fprintf(stderr, "error: no keys found for srtp\n");
    return 1;
  } else { 
    /* use gdoi keying & policy */
    /* BEW - hacked to fill in just some of the info for rtpwords app*/
    if (policy->ssrc) {
      message->header.ssrc = policy->ssrc;
    } else {
      fprintf(stderr, "error: (gdoi keying) no ssrc found\n");
      return 1;
    }
    if (policy->cipher_type != 1) {
      fprintf(stderr, "error: (gdoi keying) AES is only supported cipher\n");
      return 1;
    }
    if (policy->cipher_key_length > MAX_KEY_LEN) {
      fprintf(stderr, 
      	      "error: (gdoi keying) AES cipher len must be less than %d\n",
	      MAX_KEY_LEN);
      return 1;
    }
    memcpy(key, keys->cipher_key, policy->cipher_key_length);
    /* 
     * HACK: Check the salt length if provided. If the salt length is zero
     * then just accept the remaining amount of keying as the salt.
     */
    computed_salt_len =  MAX_KEY_LEN*2 - policy->cipher_key_length;
    if (policy->salt_key_length) {
      if (policy->salt_key_length != computed_salt_len) {
        fprintf(stderr, "error: (gdoi keying) salt len must be %d, "
		"actually is %d\n",
		policy->salt_key_length, computed_salt_len);
        return 1;
      }
    } else {
      policy->salt_key_length = computed_salt_len;
    }
    memcpy(salt, (keys->cipher_key + policy->cipher_key_length), 
    	   policy->salt_key_length);
  }
  stat = srtp_set_cipher_to_rijndael_icm(ctx, key, salt);
  if (stat)
    fprintf(stderr, "error: srtp set cipher failed with code %d\n", stat);

  return 0;
}


