/*
 * replay-database.h
 *
 * interface for a replay database for packet security
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */


#ifndef REPLAY_DB_H
#define REPLAY_DB_H

#include "integers.h"         /* for uint32_t */
#include "datatypes.h"        /* for v128_t   */


/*
 * replay_status indicates the status of a replay_check or other
 * operations involving the replay_db
 *
 * replay_check_fail_old does not necessarily indicate an attack,
 * but only that the index is earlier than the start of the
 * replay database, and hence must be assumed to not be good
 */

typedef enum {
  replay_status_ok,          /* no errors                           */
  replay_status_fail,        /* unspecified error                   */
  replay_check_ok,           /* replay check passed (good index)    */
  replay_check_fail,         /* replay check failed (bad index)     */
  replay_check_fail_old      /* replay check failed (index too old) */ 
} replay_status;


/*
 * if the ith least significant bit is one, then the packet index
 * window_end-i is in the database
 */

typedef struct {
  uint32_t window_start;   /* packet index of the first bit in bitmask */
  v128_t bitmask;  
} replay_db;

#define rdb_bits_in_bitmask (8*sizeof(v128_t))   

/*
 * replay init
 *
 * initalizes rdb
 *
 * returns replay_status_ok on success, replay_status_fail otherwise
 */

replay_status
replay_init(replay_db *rdb);


/*
 * replay_check
 *
 * checks to see if index appears in rdb
 *
 * returns replay_check_fail if the index already appears in rdb,
 * returns replay_check_ok otherwise
 */

replay_status
replay_check(const replay_db *rdb, uint32_t index);  

/*
 * replay_add_index
 *
 * adds index to replay_db (and does *not* check if index appears in db)
 *
 * returns replay_status_ok on success, replay_status_fail otherwise
 *
 */

replay_status
replay_add_index(replay_db *rdb, uint32_t index);


#endif /* REPLAY_DB_H */ 
