/*
 * ust.h
 *
 * C interface for the universal security transform
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */


#ifndef UST_H
#define UST_H
 
#include "rdbx.h"         /* defines an index and a replay database */
#include "cipher.h"       /* defines a cipher                       */
#include "auth.h"         /* defines an authentication function     */
#include "config.h"       /* for HAVE_UIO_H                         */

/* a sec_serv_t describes the security services that can be provided */

#ifndef SRTP_H

typedef enum {
  sec_serv_none          = 0,
  sec_serv_conf          = 1,
  sec_serv_auth          = 2,
  sec_serv_conf_and_auth = 3
} sec_serv_t;

#endif

/* ust_ctx_t holds the context associated with a ust connection */

typedef struct {
  cipher_t *c;            /* cipher used as keystream generator     */
  auth_t *h;              /* authentication function                */
  rdbx_t rdbx;        /* replay database                        */
  octet_t *auth_tag;        /* storage for authentication tags        */
} ust_ctx_t;

/*  ust_init(...) initializes a ust_ctx_t  */

err_status_t
ust_init(ust_ctx_t *ctx,           /* ust context                         */  
	 int index_length,         /* number of octets in the index       */
	 cipher_type_t *cipher_algo, /* keystream generator specification   */
	 octet_t *cipher_key,        /* cipher key                          */
	 int cipher_key_len,       /* number of octets in cipher key      */
	 auth_type_t *auth_algo,     /* auth algorithm specification        */
	 octet_t *auth_key,          /* auth key                            */
	 int auth_key_len,         /* number of octets in the auth key    */
	 int auth_tag_len,         /* number of octets in the auth tag    */
	 int replay_window_len     /* length of replay window (0 == none) */
	 );

/*
 * the ust_xfm function 
 *
 * ctx is the ust context, which holds the cipher and hash function
 * keys and parameters, as well as anti-replay information
 * 
 * idx is the packet index, a 48-bit unsigned integer which 
 * should be unique for each invocation of ust_xfm for a
 * given ctx
 * 
 * if auth_start != NULL, then authentication is provided to the
 * auth_len octets of data at auth_start by computing the
 * authentication tag and writing it to *tag; otherwise, the
 * authentication tag is not computed
 *
 * if enc_start != NULL, then encryption is provided to the
 * enc_len octets of data at enc_start by exoring keystream
 * into that data; otherwise, no encryption is done
 *
 * tag points to the authentication tag; after ust_xfm returns,
 * it contains the tag corresonding to the data at auth_start
 * (if auth_start != NULL) -- note that there MUST be at least
 * ust_tag_len(ctx) octets of storage at *tag! 
 */

err_status_t
ust_xfm(ust_ctx_t *ctx,    /* ust context                      */
	xtd_seq_num_t idx,  /* index                            */
	octet_t *enc_start,  /* pointer to encryption start      */
	int enc_len,       /* number of octets to encrypt      */
	octet_t *auth_start, /* pointer to authentication start  */
	int auth_len,      /* number of octets to authenticate */
	octet_t *tag         /* authentication tag               */
	);

/*
 * the ust_inv_xfm function (ust inverse transform)
 *
 * ctx is the ust context, which holds the cipher and hash function
 * keys and parameters, as well as anti-replay information
 * 
 * idx is the packet index, a 48-bit unsigned integer which 
 * should be unique for each invocation of ust_xfm for a
 * given ctx
 * 
 * if auth_start != NULL, then authentication is expected on the
 * auth_len octets of data at auth_start by computing the
 * authentication tag and comparing it to the value at *tag;
 * otherwise, no authentication check is performed
 *
 * if enc_start != NULL, then decryption is done to the enc_len
 * octets of data at enc_start by exoring keystream into that data;
 * otherwise, no decryption is done
 *
 * tag points to the authentication tag;  if auth_start != NULL,
 * then *tag is expected to hold the authentication tag corresponding
 * to the data at *auth_start -- note that there MUST be at least
 * ust_tag_len(ctx) octets of readable data at *tag! 
 */

err_status_t
ust_inv_xfm(ust_ctx_t *ctx,      /* ust context                      */
	    xtd_seq_num_t idx,         /* index                            */
	    octet_t *enc_start,  /* pointer to encryption start      */
	    int enc_len,         /* number of octets to encrypt      */
	    octet_t *auth_start, /* pointer to authentication start  */
	    int auth_len,        /* number of octets to authenticate */
	    octet_t *tag         /* authentication tag               */
	    );

/*
 * ust_tag_len(ctx) returns the length (in octets) of the
 * authentication tag for the ust context ctx.
 *
 * this function can be used to determine the storage
 * space required to hold a particular tag, if need be
 */

unsigned int 
ust_tag_len(ust_ctx_t *ctx);

/*
 * ust_xfmv is an iovec - scatter/gather interface to ust
 * 
 * this interface is not yet supported in this implementation
 */

#if HAVE_SYS_UIO_H
#include <sys/uio.h>
#else

/* iovec isn't defined in the standard headers, so define it here */
struct iovec {
  octet_t *iov_base;
  size_t iov_len;
};

#endif 

err_status_t
ust_xfmv(ust_ctx_t *ctx,         /* ust context                             */
	 xtd_seq_num_t idx,            /* index                                   */
	 struct iovec *enc_vec,  /* iovec of data to be encrypted           */
	 int enc_vec_count,      /* length of iovec of data to be encrypted */
	 struct iovec *auth_vec, /* iovec of data to be authenticated       */
	 int auth_vec_count,     /* length of iovec of data to be authntctd */
         octet_t *tag            /* authentication tag                      */
	 );

#endif /* UST_H */


