/*
 * srtp-param.x
 * 
 * secure rtp parameters, for signaling purposes
 * 
 * David A. McGrew
 * Cisco Systems, Inc.
 */

/* cipher_id identifies a particular type of cipher */

enum cipher_id {
  null_cipher = 0,
  aes_cm      = 1, 
  aes_f8      = 2  
};

/*
 * cipher_descr defines a particular choice of parameters for a
 * particular cipher
 */

struct cipher_descr {
  cipher_id type;
  unsigned int key_length;
};

/* auth_func_id identifies a particular type of authentication function */

enum auth_func_id {
  null_auth_func = 0,
  tmmhv2         = 1,
  hmac_sha1      = 2
};

/*
 * auth_params contains a particular choice of parameters for
 * authentication
 */

struct auth_params {
  unsigned int key_length;
  unsigned int tag_length;
  unsigned int replay_window_size; 
};

/*
 *  an auth_descr describes a choice of parameters for providing
 *  message authentication
 */

union auth_descr switch (auth_func_id type) {
 case null_auth_func:      
   void;                  /* null auth., no need for parameters */
 default:
   auth_params auth;
};

/* enc_options describes the encryption options */

enum enc_options {
  no_encryption                = 0,   
  encryption                   = 1,
  encrypt_with_forward_secrecy = 2    /* not yet supported */
};

/* packet_index is a 48-bit integer associated with an srtp packet */

struct packet_index {
  unsigned int roc;  /* rollover counter (highest 32 bits) */
  unsigned int seq;  /* sequence number (lowest 16 bits)   */
};

struct srtp_parameters {
  unsigned int ssrc;                  /* rtp sender source         */
  cipher_descr cipher;                /* keystream generator       */
  enc_options  srtp_enc_opt;          /* encryption options        */
  enc_options  srtcp_enc_opt;         /* srtcp encryption options  */
  auth_descr   auth;                  /* authentication parameters */
  octet        key<>;                 /* master key                */
  octet        salt<>;                /* master salt               */
  unsigned int key_derivation_rate;   /* re-derivation rate        */  
  packet_index from;                  /* first ok packet_index     */
  packet_index to;                    /* last ok packet_index      */ 
}; 

/*
 * an ust_descr describes a choice of ust parameters
 */

struct ust_descr {
  cipher_descr cipher;  /* the keystream generator                       */
  auth_descr   auth;    /* if null, don't provide message authentication */
  enc_options enc_opt;  /* encryption is optional                        */
};

/*
 * a cipher_definition or auth_func_definition defines the interface to and
 * the metadata associated with a particular cipher or auth_func function
 */

struct cipher_definition {
  cipher_id type;               /* type identifier                       */
  string key_length<16>;        /* list of valid key lengths in octets   */
  unsigned int index_length;    /* the index length in octets            */
  unsigned int segment_length;  /* max. keystream segment len. in octets */ 
  string description<256>;      /* human-readable description            */
};

struct auth_func_definition {
  auth_func_type type;          /* type identifier                        */
  string key_length<16>;        /* list of valid key lengths in octets    */
  string output_length<16>;     /* list of valid output lengths in octets */
  unsigned int input_length;    /* maximum input length in octets         */ 
  string description<256>;      /* human-readable description             */
};

/*
 * ust provides the following service options:
 *
 * confidentiality
 * (uses a keystream generator)
 *
 * confidentiality and message authentication
 * (uses a keystream generator and auth_func)
 *
 * confidentiality, message authentication, and replay protection
 * (uses a keystream generator and auth_func)
 *
 * (confidentiality optionally provides forward security)
 * 
 */




   

   
