/*
 * srtp.c
 *
 *  the secure real-time transport protocol
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */

/*
 *	
 * Copyright  (c) 2001, 2002, Cisco Systems, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.
 * 
 *   Neither the name of the Cisco Systems, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdlib.h>          /* for malloc() */
#include "srtp.h"
#include "rijndael-icm.h"    /* for rijndael_icm                 */

extern cipher_type_t rijndael_icm;


#define PRINT_DEBUG 0      /* set to 1 to print debugging info */
#define PRINT_DEBUG_AUTH 0 /* set to 1 to print auth debugging info */

#if (PRINT_DEBUG || PRINT_DEBUG_AUTH)
#include <stdio.h>
#endif

#define octets_in_rtp_header  12
#define uint32s_in_rtp_header 3

err_status_t
srtp_protect(srtp_ctx_t *ctx, srtp_hdr_t *hdr, int *pkt_octet_len) {
  uint32_t *enc_start;      /* pointer to start of encrypted portion  */
  int enc_octet_len;        /* number of octets in encrypted portion  */
  xtd_seq_num_t est;        /* estimated xtd_seq_num_t of *hdr        */
  int delta;                /* delta of local pkt idx and that in hdr */
  octet_t *auth_tag = NULL; /* location of auth_tag within packet     */
  int status;            
  int tag_len = ctx->authenticator->out_len;

  /*
   * find starting point for encryption - the encrypted portion starts
   * after the rtp header extension, if present; otherwise, it starts
   * after the last csrc, if any are present
   */
  enc_start = (uint32_t *)hdr + uint32s_in_rtp_header + hdr->cc;  
  if (hdr->x == 1) 
    enc_start += ((srtp_hdr_xtnd_t *)enc_start)->length;

  /*
   * estimate the packet index using the start of the replay window   
   * and the sequence number from the header
   */
  delta = rdbx_estimate_index(&ctx->rdbx, &est, ntohs(hdr->seq));
  if (delta > 0)
    rdbx_add_index(&ctx->rdbx, delta);
  else
    return err_status_replay_fail;  /* we've been asked to reuse an index */

  /* if we're using rindael counter mode, exor the ssrc into the salt */
  if (ctx->encryptor->type == &rijndael_icm) {
    uint32_t ssrc = ntohl(hdr->ssrc);
    rijndael_icm_context *cipher = (rijndael_icm_context *)ctx->encryptor;

    /* exor the ssrc into bytes four through seven of the salt */
    cipher->offset.octet[4] ^= (ssrc >> 24);
    cipher->offset.octet[5] ^= (ssrc >> 16) & 0xff;
    cipher->offset.octet[6] ^= (ssrc >> 8) & 0xff;
    cipher->offset.octet[7] ^= ssrc & 0xff;
  }

  /* set the cipher to the segment of the estimated packet index */
  status = cipher_set_segment(ctx->encryptor, est);
  if (status)
    return status;

  /* if we're using rindael counter mode, exor the ssrc out of the salt */
  if (ctx->encryptor->type == &rijndael_icm) {
    uint32_t ssrc = ntohl(hdr->ssrc);
    rijndael_icm_context *cipher = (rijndael_icm_context *)ctx->encryptor;

    /* exor the ssrc into bytes four through seven of the salt */
    cipher->offset.octet[4] ^= (ssrc >> 24);
    cipher->offset.octet[5] ^= (ssrc >> 16) & 0xff;
    cipher->offset.octet[6] ^= (ssrc >> 8) & 0xff;
    cipher->offset.octet[7] ^= ssrc & 0xff;
  }

#if PRINT_DEBUG
  printf("ssrc:       00000000%s0000000000000000\n", 
	 octet_string_hex_string((octet_t *)&ssrc, 8));  
  printf("c slt^ssrc: %s\n", 
	 octet_string_hex_string(enc_salt_buf, 32));  
#endif


  /* 
   * set the tag value to the keystream prefix if authentication is
   * needed
   */
  if (ctx->services & sec_serv_auth) {
    auth_tag = (octet_t *)hdr + *pkt_octet_len;
    octet_string_set_to_zero(auth_tag, tag_len);
#if PRINT_DEBUG_AUTH
    printf("srtp_protect tag and following four octets:            %s\n", 
	   octet_string_hex_string(auth_tag, 2*tag_len + 8));
#endif

    status = cipher_encrypt(ctx->encryptor, auth_tag, tag_len);
    if (status)
      return status;
#if PRINT_DEBUG_AUTH
    printf("srtp_protect tag (k prefix) and following four octets: %s\n", 
	   octet_string_hex_string(auth_tag, 2*tag_len + 8));
#endif
  }

  /* encrypt the data part of the packet, if needed */
  if (ctx->services & sec_serv_conf) {
    enc_octet_len = *pkt_octet_len - ((enc_start - (uint32_t *)hdr) << 2);
    status = cipher_encrypt(ctx->encryptor, 
			    (octet_t *)enc_start, enc_octet_len);
  }
  /*    printf("after encryption:\n"); */
  /*    srtp_print_packet(hdr, *pkt_octet_len); */

  /*
   * compute the hash and add value into the authentication tag, if
   * authentication needed
   */
  if (ctx->services & sec_serv_auth) {
    status = auth_compute(ctx->authenticator, (octet_t *)hdr,
			  *pkt_octet_len, auth_tag);
    if (status)
      return status;
    
#if PRINT_DEBUG_AUTH
    printf("srtp_protect tag and following four octets:            %s\n", 
	   octet_string_hex_string(auth_tag, 2*tag_len + 8));
#endif
    
    /* increase the packet length by the length of the auth tag */
    *pkt_octet_len += tag_len;
  }

  return err_status_ok;  
}

err_status_t
srtp_unprotect(srtp_ctx_t *ctx, srtp_hdr_t *hdr, int *pkt_octet_len) {
  uint32_t *enc_start;   /* pointer to start of encrypted portion */
  int enc_octet_len;     /* number of octets in encrypted portion */
  xtd_seq_num_t est;     /* estimated xtd_seq_num_t of *hdr       */
  int delta;             /* difference between est and the        *
                          * xtd_seq_num_t in the replay database  */
  int status;
  octet_t *msg_tag;      /* location of auth_tag in packet        */
  int tag_len = ctx->authenticator->out_len;

  /* estimate packet index from seq. num. in header */
  delta = rdbx_estimate_index(&ctx->rdbx, &est, ntohs(hdr->seq));

  /* check replay database */
  if (rdbx_check(&ctx->rdbx, delta) != replay_check_ok)
    return err_status_replay_fail;

  /* if we're using rindael counter mode, exor the ssrc into the salt */
  if (ctx->encryptor->type == &rijndael_icm) {
    uint32_t ssrc = ntohl(hdr->ssrc);
    rijndael_icm_context *cipher = (rijndael_icm_context *)ctx->encryptor;

    /* exor the ssrc into bytes four through seven of the salt */
    cipher->offset.octet[4] ^= (ssrc >> 24);
    cipher->offset.octet[5] ^= (ssrc >> 16) & 0xff;
    cipher->offset.octet[6] ^= (ssrc >> 8) & 0xff;
    cipher->offset.octet[7] ^= ssrc & 0xff;
  }

  /* set the cipher to the proper keystream segment */
  status = cipher_set_segment(ctx->encryptor, est);
  if (status)
    return err_status_cipher_fail;

  /* if we're using rindael counter mode, exor the ssrc out of the salt */
  if (ctx->encryptor->type == &rijndael_icm) {
    uint32_t ssrc = ntohl(hdr->ssrc);
    rijndael_icm_context *cipher = (rijndael_icm_context *)ctx->encryptor;

    /* exor the ssrc into bytes four through seven of the salt */
    cipher->offset.octet[4] ^= (ssrc >> 24);
    cipher->offset.octet[5] ^= (ssrc >> 16) & 0xff;
    cipher->offset.octet[6] ^= (ssrc >> 8) & 0xff;
    cipher->offset.octet[7] ^= ssrc & 0xff;
  }

  /*
   * if authentication is needed, generate the keystream prefix, hash
   * the message, then compare tags
   */
  if (ctx->services & sec_serv_auth) {
    /* set the tag value to the keystream prefix */
    octet_string_set_to_zero(ctx->auth_tag, tag_len);
    status = cipher_encrypt(ctx->encryptor, ctx->auth_tag, tag_len);
    if (status)
      return status;

#if PRINT_DEBUG_AUTH
  printf("srtp_unprotect tag (k prefix):          %s\n", 
	 octet_string_hex_string(ctx->auth_tag, 2*tag_len));
#endif
  
  /* compute hash of packet and add that value into the auth_tag */
  status = auth_compute(ctx->authenticator, (octet_t *)hdr,
			(*pkt_octet_len) - tag_len, ctx->auth_tag);
  if (status) 
    return err_status_auth_fail;
  
  /*
   * the auth tag in the packet starts at tag_len octets from the end
   * of the packet
   */
  msg_tag = (octet_t *)hdr + *pkt_octet_len - tag_len;
  
#if PRINT_DEBUG_AUTH
  printf("srtp_unprotect tag:          %s\n", 
	 octet_string_hex_string(ctx->auth_tag, 2*tag_len));
  printf("srtp_unprotect tag (packet): %s\n", 
	 octet_string_hex_string(msg_tag, 2*tag_len));
#endif
    
    /* compare computed auth tag with that in the packet */
    if (octet_string_is_eq(ctx->auth_tag, msg_tag, tag_len)) 
      return err_status_auth_fail;
    
    /* update pkt_octet_len */
    *pkt_octet_len -= tag_len;

  }
  
  /*
   * add the index to rdb and set the local sequence number and
   * rollover counter
   */
  rdbx_add_index(&ctx->rdbx, delta);  /* return value not checked, for now */

  
  if (ctx->services & sec_serv_conf) {
    /*
     * find starting point for encryption - the encrypted portion starts
     * after the rtp header extension, if present; otherwise, it starts
     * after the last csrc, if any are present
     */
    enc_start = (uint32_t *)hdr + uint32s_in_rtp_header + hdr->cc;
    
    if (hdr->x == 1) 
      enc_start += ((srtp_hdr_xtnd_t *)enc_start)->length;
    
    /* decrypt packet using estimated packet index */
    enc_octet_len = (enc_start - (uint32_t *)hdr) << 2;
    status = cipher_encrypt(ctx->encryptor, (octet_t *)enc_start,
			    *pkt_octet_len - enc_octet_len);

    if (status)
      return status;
  }
  return err_status_ok;
}

err_status_t
srtp_alloc(srtp_ctx_t **ctx_ptr,
	   cipher_type_t *ctype,
	   int cipher_key_len, 
	   auth_type_t *atype,
	   int auth_key_len,
	   int auth_tag_len,
	   sec_serv_t sec_serv) {
  err_status_t stat;
  extern auth_type_t null_auth;
  srtp_ctx_t *ctx;

  /* allocate srtp context and set ctx_ptr */
  ctx = (srtp_ctx_t *) malloc(sizeof(srtp_ctx_t));
  if (ctx == NULL)
    return err_status_alloc_fail;
  *ctx_ptr = ctx;

  /* initialize replay database */
  rdbx_init(&ctx->rdbx);

  /* initialize cipher */
  stat = cipher_type_alloc(ctype, &ctx->encryptor, cipher_key_len);
  if (stat) {
    free(ctx);
    return stat;
  }
      
  /* initalize authentication function */
  stat = auth_type_alloc(atype, &ctx->authenticator,
			 auth_key_len, auth_tag_len);
  if (stat) {
    free(ctx);
    cipher_dealloc(ctx->encryptor);
    return stat;
  }

  /* 
   * if auth is not null_auth, then allocate storage area for
   * authentication tag; otherwise, set auth_tag to null
   */
  if (atype != &null_auth)
    ctx->auth_tag = malloc(auth_get_tag_length(ctx->authenticator));
  else
    ctx->auth_tag = NULL;

  /* set security service flag */
  ctx->services = sec_serv;

  return err_status_ok;
}

err_status_t
srtp_dealloc(srtp_ctx_t *ctx) {

  /* deallocate storage for cipher, auth func, and auth tag */
  cipher_dealloc(ctx->encryptor);
  auth_dealloc(ctx->authenticator);
  free(ctx->auth_tag);
  free(ctx);

  return err_status_ok;
}



/*
 * srtp_init_aes_128_prf(ctx, key, salt) initializes an srtp_ctx_t
 * with a given master key and master salt.
 * 
 * the key MUST be sixteen octets long
 *
 * the salt MUST be fourteen octets long, though the last two octets
 * will be ignored
 *
 * The PRF used for key derivation here is that defined in the secure
 * rtp specification, though in theory, other PRFs can be used.  The
 * cipher is hardwired to AES-128 Counter Mode with 112 bits of salt.
 */

typedef enum {
  label_encryption             = 0x00,
  label_message_authentication = 0x01,
  label_salt                   = 0x02
} srtp_prf_label;

err_status_t
srtp_init_aes_128_prf(srtp_ctx_t *srtp, const octet_t key[16],
		      const octet_t salt[14]) {
  err_status_t stat;
  rijndael_icm_context c;
  xtd_seq_num_t idx = { 0, 0 };        /* for setting icm to zero-index   */   
  octet_t *buffer;                     /* temporary storage for keystream */
  octet_t *enc_key_buf, *enc_salt_buf, *auth_key_buf; 
  int buffer_size;
  octet_t initial_counter[16];

#if PRINT_DEBUG
  printf("srtp_init_aes_128_prf()\n");
#endif
  /* set intitial_counter */
  initial_counter[0]  = salt[0];
  initial_counter[1]  = salt[1];
  initial_counter[2]  = salt[2];
  initial_counter[3]  = salt[3];
  initial_counter[4]  = salt[4];
  initial_counter[5]  = salt[5];
  initial_counter[6]  = salt[6];
  initial_counter[7]  = salt[7];
  initial_counter[8]  = salt[8];
  initial_counter[9]  = salt[9];
  initial_counter[10] = salt[10];
  initial_counter[11] = salt[11];
  initial_counter[12] = salt[12];
  initial_counter[13] = salt[13];
  initial_counter[14] = 0x00;
  initial_counter[15] = 0x00;

  /* allocate temporary storage and set pointers */
  buffer_size = cipher_get_key_length(srtp->encryptor)
    + cipher_get_salt_length(srtp->encryptor) 
    + auth_get_key_length(srtp->authenticator);
  buffer = malloc(buffer_size);
  if (buffer == NULL)
    return err_status_alloc_fail;
  enc_key_buf = buffer;
  enc_salt_buf = buffer + cipher_get_key_length(srtp->encryptor);
  auth_key_buf = enc_salt_buf + cipher_get_salt_length(srtp->encryptor);
    
  /* generate encryption key, putting it into enc_key_buf  */

  /* note that we assume that index DIV t == 0 in this implementation */
  /* exor <label> into the eigth octet of the state */
  initial_counter[7] = salt[7] ^ label_encryption;
  
  rijndael_icm_context_init(&c, key, initial_counter);
  rijndael_icm_set_segment(&c, idx);

#if PRINT_DEBUG
  printf("master key: %s\n", 
	 octet_string_hex_string((octet_t *)&c.expanded_key[0], 32));
  printf("master slt: %s\n", 
	 octet_string_hex_string(salt, 28));
  printf("generating cipher key\n");
  printf("cipher ctr: %s\n", 
	 octet_string_hex_string((octet_t *)&c.counter, 32));  
#endif

  rijndael_icm_encrypt(&c, enc_key_buf, 
		       cipher_get_key_length(srtp->encryptor));

#if PRINT_DEBUG
  printf("cipher key: %s\n", 
	 octet_string_hex_string(enc_key_buf, 32));  
#endif

  /* generate encryption salt, putting it into enc_salt_buf */
  /* exor <label> into the eigth octet of the state */
  initial_counter[7] = salt[7] ^ label_salt;
 
  rijndael_icm_context_init(&c, key, initial_counter);
  rijndael_icm_set_segment(&c, idx);

#if PRINT_DEBUG
  printf("generating cipher salt\n");
  printf("cr slt ctr: %s\n", 
	 octet_string_hex_string((octet_t *)&c.counter, 32));  
#endif

  rijndael_icm_encrypt(&c, enc_salt_buf, 
		       cipher_get_salt_length(srtp->encryptor));

#if PRINT_DEBUG
  printf("cipher slt: %s\n", 
	 octet_string_hex_string(enc_salt_buf, 32));  
#endif

  /*
   * we don't yet know the ssrc of the sender, so we don't exor the
   * ssrc value into the enc_salt_buf
   */

  /* initialize cipher */
  stat = cipher_init(srtp->encryptor, enc_key_buf, enc_salt_buf); 
  if (stat) {
    free(buffer);
    return err_status_alloc_fail;
  }
    
  /* generate authentication key, putting it into auth_key_buf */
  /* exor <label> into the eigth octet of the state */
  initial_counter[7] = salt[7] ^ label_message_authentication;

  rijndael_icm_context_init(&c, key, initial_counter);
  rijndael_icm_set_segment(&c, idx);

#if PRINT_DEBUG
  printf("generating auth key\n");
  printf("auth ctr:   %s\n",
	 octet_string_hex_string((octet_t *)&c.counter, 32));  
#endif

  rijndael_icm_encrypt(&c, auth_key_buf,
		       auth_get_key_length(srtp->authenticator));
#if PRINT_DEBUG
  printf("auth key:   %s\n",
	 octet_string_hex_string(auth_key_buf, 
				 2*auth_get_key_length(srtp->authenticator))); 
  printf("ssrc: %x\n", ssrc);
#endif
  
  /* initialize authentication function */
  auth_init(srtp->authenticator, auth_key_buf,
	    auth_get_key_length(srtp->authenticator));
  if (stat) {
    free(buffer);
    return err_status_init_fail;
  }  

  /* free memory then return */
  free(buffer);
  return err_status_ok;  
}


/*
 * srtp_get_trailer_length(&a) returns the number of octets that will
 * be added to an RTP packet by the SRTP processing.  This value
 * is constant for a given srtp_ctx_t (i.e. between initializations).
 */

int
srtp_get_trailer_length(const srtp_t a) {
  return auth_get_tag_length(a->authenticator);
}

/* 
 * srtp_print_packet(...) is for debugging only 
 * it prints an RTP packet to the stdout
 */

#include <stdio.h>

void
srtp_print_packet(srtp_hdr_t *hdr, int pkt_octet_len) {
  octet_t *data = ((octet_t *)hdr)+octets_in_rtp_header;
  int hex_len = 2*(pkt_octet_len-octets_in_rtp_header);

  printf("rtp packet: {\n");
  printf("   version:\t%d\n", hdr->version);
  printf("   p:\t\t%d\n", hdr->p);
  printf("   x:\t\t%d\n", hdr->x);
  printf("   cc:\t\t%d\n", hdr->cc);
  printf("   m:\t\t%d\n", hdr->m);
  printf("   pt:\t\t%x\n", hdr->pt);
  printf("   seq:\t\t%x\n", hdr->seq);
  printf("   ts:\t\t%x\n", hdr->ts);
  printf("   ssrc:\t%x\n", hdr->ssrc);
  printf("   data:\t%s\n", octet_string_hex_string(data, hex_len));
  printf("} (%d octets in data)\n", pkt_octet_len-octets_in_rtp_header);

}




