/*
 * replay-driver.c
 *
 * A driver for the replay_database implementation
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */

/*
 *	
 * Copyright  (c) 2001, 2002, Cisco Systems, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.
 * 
 *   Neither the name of the Cisco Systems, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>

#include "rdb.h"
#include "ut-sim.h"

void
test_replay_db();

double
replay_check_adds_per_second();

int
main () {

  printf("testing replay_database...\n");
  test_replay_db();
  printf("done\n");

  printf("replay_check/replay_adds per second: %e\n",
	 replay_check_adds_per_second());
  
  return 0;
}


void
print_rdb(replay_db *rdb) {
  printf("rdb: {%u, %s}\n", rdb->window_start, v128_bit_string(rdb->bitmask));
}

void
replay_check_add(replay_db *rdb, uint32_t idx) {

  if (replay_check(rdb, idx) != replay_check_ok) {
    printf("replay_check failed at index %u\n", idx);
    return;
  }
  if (replay_add_index(rdb, idx) != replay_status_ok)
    printf("replay_add_index failed at index %u\n", idx);

}

void
replay_check_expect_failure(replay_db *rdb, uint32_t idx) {

  if (replay_check(rdb, idx) != replay_check_fail)
    printf("replay_check failed at index %u (false positive)\n", idx);
}

void
replay_check_unordered(replay_db *rdb, uint32_t idx) {
  replay_status rstat;

  rstat = replay_check(rdb, idx);
  if ((rstat != replay_check_ok) && (rstat != replay_check_fail_old))
    printf("replay_check_unordered failed at index %u\n", idx);
}



#define MAX_IDX 160

void
test_replay_db() {
  replay_db rdb;
  uint32_t idx, ircvd;
  ut_connection utc;
  
  if (replay_init(&rdb) != replay_status_ok) {
    printf("replay_init failed\n");
    exit(1);
  }

  /* test sequential insertion */
  for (idx=0; idx < MAX_IDX; idx++) {
    replay_check_add(&rdb, idx);
  }

  /* test for false positives */
  for (idx=0; idx < MAX_IDX; idx++) {
    replay_check_expect_failure(&rdb, idx);
  }

  /* re-initialize */
  if (replay_init(&rdb) != replay_status_ok) {
    printf("replay_init failed\n");
    exit(1);
  }

  /* test non-sequential insertion */
  ut_init(&utc);
  
  for (idx=0; idx < MAX_IDX; idx++) {
    ircvd = ut_next_index(&utc);
    replay_check_unordered(&rdb, ircvd);
  }
  
}

#include <time.h>       /* for clock()  */
#include <stdlib.h>     /* for random() */

#define REPLAY_NUM_TRIALS 10000000

double
replay_check_adds_per_second() {
  uint32_t i;
  replay_db rdb;
  clock_t timer;
  int failures;                    /* count number of failures        */
  
  if (replay_init(&rdb) != replay_status_ok) {
    printf("replay_init failed\n");
    exit(1);
  }  

  timer = clock();
  for(i=0; i < REPLAY_NUM_TRIALS; i+=3) {
    if (replay_check(&rdb, i+2) != replay_check_ok)
      ++failures;
    if (replay_add_index(&rdb, i+2) != replay_status_ok)
      ++failures;
    if (replay_check(&rdb, i+1) != replay_check_ok)
      ++failures;
    if (replay_add_index(&rdb, i+1) != replay_status_ok)
      ++failures;
    if (replay_check(&rdb, i) != replay_check_ok)
      ++failures;
    if (replay_add_index(&rdb, i) != replay_status_ok)
      ++failures;
  }
  timer = clock() - timer;

  return (double) CLOCKS_PER_SEC * REPLAY_NUM_TRIALS / timer;
}
