/*
 * srtp-driver.c
 * 
 * a test driver for the srtp implementation
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */

/*
 *	
 * Copyright  (c) 2001, 2002, Cisco Systems, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.
 * 
 *   Neither the name of the Cisco Systems, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <time.h>     /* for clock()           */
#include <stdlib.h>   /* for random ()         */
#include <stdio.h>    /* for print(), fflush() */
#include <unistd.h>   /* for getopt()          */

#include "srtp.h"

#define SRTP_VERBOSE 0   /* set to 1 to print packets to stdout */

extern cipher_type_t rijndael_icm;
extern auth_type_t   tmmhv2;

double
srtp_bits_per_second(int msg_len_octets);

double
srtp_rejections_per_second(int msg_len_octets);

err_status_t
srtp_test();

err_status_t
srtp_test_prf();

void
usage(char *prog_name) {
  printf("usage: %s [ -t | -v ]\n", prog_name);
  exit(255);
}

int
main (int argc, char *argv[]) {
  int len;
  char q;
  unsigned do_timing_test = 0;
  unsigned do_validation = 0;

  /* process input arguments */
  while (1) {
    q = getopt(argc, argv, "tv");
    if (q == -1) 
      break;
    switch (q) {
    case 't':
      do_timing_test = 1;
      break;
    case 'v':
      do_validation = 1;
      break;
    default:
      usage(argv[0]);
    }    
  }
 
  printf("srtp driver\nDavid A. McGrew\nCisco Systems, Inc.\n");

  if (!do_validation && !do_timing_test)
    usage(argv[0]);

  if (do_validation) {
    printf("testing srtp_protect and srtp_unprotect...\n");
    fflush(stdout);
    if (srtp_test() == err_status_ok) {
      printf("\ttest passed\n");
    } else {
      printf("\ttest failed\n");
      exit(1);
    }
    if (srtp_test_prf() == err_status_ok) {
      printf("\ttest_prf passed\n");
    } else {
      printf("\ttest_prf failed\n");
      exit(1);
    }
    printf("done\n");
  }
  
  if (do_timing_test) {
    printf("testing srtp throughput:\n");
    printf("message length (octets)\tthroughput (megabits per second)\n");
    
    for (len=8; len < 512; len *= 2)
      printf("%d\t\t\t%f\n", len, srtp_bits_per_second(len+12) / 1.0E6);
    
    printf("testing srtp rejection throughput:\n");
    printf("message length (octets)\trejections per second\n");
    
    for (len=8; len < 512; len *= 2)
      printf("%d\t\t\t%e\n", len, srtp_rejections_per_second(len));
  }
  
  return 0;  
}

/* test functions for srtp_ctx_t */

#include <assert.h>


/*
 * srtp_construct_test_packet(len) returns a pointer to a
 * (malloced) example RTP packet of length len.  The four
 * octets following the packet are filled with 0xff values
 * to enable testing for overwrites.
 *
 * note that the location of the test packet can (and should) be
 * deallocated with the free() call once it is no longer needed.
 */

srtp_hdr_t *
srtp_construct_test_packet(int pkt_octet_len) {
  int i;
  uint32_t *buffer;
  srtp_hdr_t *hdr;
    
  assert(sizeof(srtp_hdr_t) == 12);

  /* allocate memory for test packet */
  hdr = malloc(pkt_octet_len + sizeof(srtp_hdr_t) + 4);
  if (!hdr)
    return NULL;
  
  hdr->version = 2;         /* RTP version two     */
  hdr->p    = 0;            /* no padding needed   */
  hdr->x    = 0;            /* no header extension */
  hdr->cc   = 0;            /* no CSRCs            */
  hdr->m    = 0;            /* marker bit          */
  hdr->pt   = 0xf;          /* payload type        */
  hdr->seq  = 0x1234;       /* sequence number     */
  hdr->ts   = 0xffffffff;   /* timestamp           */
  hdr->ssrc = 0xcafebabe;   /* synch. source       */

  buffer = (uint32_t *)hdr;
  buffer += 3;
  pkt_octet_len -= 12;

  /* set RTP data to C5C0 */
  for (i=0; i < pkt_octet_len/4; i++)
    *buffer++ = 0xc5c0c5c0;
  /* set post-data value to 0xffff to enable overrun checking */
  *buffer = 0xffffffff;

  return hdr;
}

#define MAX_MSG_LEN 1024

double
srtp_bits_per_second(int msg_len_octets) {
  srtp_ctx_t *srtp;
  srtp_hdr_t *mesg;  
  int i;
  clock_t timer;
  int num_trials = 100000;
  int len;
  unsigned char test_key[16] = {
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf
  };
  unsigned char salt[16] = {
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf
  };
  const uint16_t tmmhv2_key1[47] = {
  0xe627, 0x6a01, 0x5ea7, 0xf27a, 0xc536, 0x2192, 0x11be, 0xea35,
  0xdb9d, 0x63d6, 0xfa8a, 0xfc45, 0xe08b, 0xd216, 0xced2, 0x7853,
  0x1a82, 0x22f5, 0x90fb, 0x1c29, 0x708e, 0xd06f, 0x82c3, 0xbee6,
  0x4f21, 0x6f33, 0x65c0, 0xd211, 0xc25e, 0x9138, 0x4fa3, 0x7c1f,
  0x61ac, 0x3489, 0x2976, 0x8c19, 0x8252, 0xddbf, 0xcad3, 0xc28f,
  0x68d6, 0x58dd, 0x504f, 0x2bbf, 0x0278, 0x70b7, 0xcfca
  };
  err_status_t status;

  status = srtp_alloc(&srtp,         /* srtp context                */
	     &rijndael_icm,	     /* cipher type                 */
	     16,		     /* cipher key length in octets */
	     &tmmhv2,		     /* authentication func type    */
	     94,		     /* auth key length in octets   */
	     4,			     /* auth tag length in octets   */
	     sec_serv_conf_and_auth  /* security services flag      */
	     );
  if (status)
    return status;

  cipher_init(srtp->encryptor, test_key, salt);
  auth_init(srtp->authenticator, (octet_t *)tmmhv2_key1, 94);
	    
  mesg = srtp_construct_test_packet(msg_len_octets
				    + srtp_get_trailer_length(srtp));
  if (mesg == NULL)
    return 0.0;   /* indicate failure by returning zero */
  
  timer = clock();
  for (i=0; i < num_trials; i++) {
    len = msg_len_octets;
    srtp_protect(srtp, (srtp_hdr_t *)mesg, &len);
  }
  timer = clock() - timer;

  free(mesg);
  
  return (double) (msg_len_octets - 12) * 8 *
                  num_trials * CLOCKS_PER_SEC / timer;   
}

double
srtp_rejections_per_second(int msg_len_octets) {
  srtp_ctx_t *srtp;
  srtp_hdr_t *mesg; 
  int i;
  int len;
  clock_t timer;
  int num_trials = 1000000;
  unsigned char test_key[16] = {
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf
  };
  unsigned char salt[16] = {
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf
  };
  unsigned char tmmhv2_key[94] = { 0xff, };

  srtp_alloc(&srtp,                  /* srtp context                */
	     &rijndael_icm,	     /* cipher type                 */
	     16,		     /* cipher key length in octets */
	     &tmmhv2,		     /* authentication func type    */
	     94,		     /* auth key length in octets   */
	     4,			     /* auth tag length in octets   */
	     sec_serv_conf_and_auth  /* security services flag      */
	     );
  cipher_init(srtp->encryptor, test_key, salt);
  auth_init(srtp->authenticator, tmmhv2_key, 94);

  mesg = srtp_construct_test_packet(msg_len_octets
				    + srtp_get_trailer_length(srtp));
  
  len = msg_len_octets;  
  srtp_protect(srtp, (srtp_hdr_t *)mesg, &len);
  
  timer = clock();
  for (i=0; i < num_trials; i++) {
    len = msg_len_octets;
    srtp_unprotect(srtp, (srtp_hdr_t *)mesg, &len);
  }
  timer = clock() - timer;

  free(mesg);
  
  return (double) num_trials * CLOCKS_PER_SEC / timer;   
}


void
srtp_err(err_status_t s) {
  switch(s) {
  case err_status_ok:
    return;
  case err_status_auth_fail:
    fprintf(stderr, "srtp authentication failure\n");
  default:
    fprintf(stderr, "srtp failure\n");
  }
  exit (1);
}

#include <string.h>
  
err_status_t
srtp_test() {
  int i;
  srtp_ctx_t *srtp_sender, *srtp_rcvr;
  err_status_t status = err_status_ok;
  srtp_hdr_t *hdr, *hdr2;
  octet_t *pkt_end;
  int msg_len_octets;
  int len;
  unsigned char test_key[16] = {
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf
  };
  unsigned char salt[16] = {
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0x0, 0x0
  };
  unsigned char tmmhv2_key[94] = { 
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
    0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
    0x0, 0x1, 0x2, 0x3
  };
  	 
  srtp_err(srtp_alloc(&srtp_sender,          /* srtp context                */
		      &rijndael_icm,	     /* cipher type                 */
		      16,		     /* cipher key length in octets */
		      &tmmhv2,		     /* authentication func type    */
		      94,		     /* auth key length in octets   */
		      4,		     /* auth tag length in octets   */
		      sec_serv_conf_and_auth /* security services flag      */
	     ));
  cipher_init(srtp_sender->encryptor, test_key, salt);
  auth_init(srtp_sender->authenticator, (octet_t *)tmmhv2_key, 94); 

  /* initialize data buffer */
  msg_len_octets = 28;
  hdr  = srtp_construct_test_packet(msg_len_octets
				    + srtp_get_trailer_length(srtp_sender));
  if (hdr == NULL)
    return err_status_alloc_fail;
  hdr2 = srtp_construct_test_packet(msg_len_octets
				    + srtp_get_trailer_length(srtp_sender)); 
  if (hdr2 == NULL) {
    free(hdr);
    return err_status_alloc_fail;
  }
  
  /* set message length */
  len = msg_len_octets;

  /* set pkt_end (for later checking for overruns) */
  pkt_end = (octet_t *)hdr +
    msg_len_octets + srtp_get_trailer_length(srtp_sender);
  
#if SRTP_VERBOSE
  printf("before protection:\npacket length: %d\n", len);
  srtp_print_packet(hdr, len);
#endif

  srtp_err(srtp_protect(srtp_sender, hdr, &len));
  
#if SRTP_VERBOSE
  printf("after protection:\npacket length: %d\n", len);
  srtp_print_packet(hdr, len);
#endif

  /* check for overrun of the srtp_protect() function */
  for (i = 0; i < 4; i++)
    if (pkt_end[i] != 0xff) {
      fprintf(stdout, "overwrite in srtp_protect() function "
	      "(expected %x, found %x in trailing octet %d)\n",
	      0xff, ((octet_t *)hdr)[i], i);
      free(hdr);
      free(hdr2);
      return err_status_algo_fail;
    }
  
  srtp_err(srtp_alloc(&srtp_rcvr,            /* srtp context                */
		      &rijndael_icm,	     /* cipher type                 */
		      16,		     /* cipher key length in octets */
		      &tmmhv2,		     /* authentication func type    */
		      94,		     /* auth key length in octets   */
		      4,		     /* auth tag length in octets   */
		      sec_serv_conf_and_auth /* security services flag      */
	     ));
  cipher_init(srtp_rcvr->encryptor, test_key, salt);
  auth_init(srtp_rcvr->authenticator, (octet_t *)tmmhv2_key, 94);
  
  srtp_err(srtp_unprotect(srtp_rcvr, hdr, &len));

#if SRTP_VERBOSE
  printf("after unprotection:\npacket length: %d\n", len);
  srtp_print_packet(hdr, len);
#endif
  fflush(stdout);

  for (i=0; i < msg_len_octets; i++)
    if (((octet_t *)hdr)[i] != ((octet_t *)hdr2)[i]) {
      fprintf(stdout, "mismatch at octet %d\n", i);
      fprintf(stdout, "expected %d, got %d\n",
	      ((octet_t *)hdr2)[i], ((octet_t *)hdr)[i]);
      status = err_status_algo_fail;
    }

#if SRTP_VERBOSE
  printf("reference packet:\npacket length: %d\n", len);
  srtp_print_packet(hdr2, len);
#endif
  
  /* DAM - in the future, test for false negatives here */  

  return status;
}


err_status_t
srtp_test_prf() {
  int i;
  srtp_ctx_t *srtp_sender, *srtp_rcvr;
  err_status_t status = err_status_ok;
  srtp_hdr_t *hdr, *hdr2;
  octet_t *pkt_end;
  int msg_len_octets;
  int len;
  unsigned char test_key[16] = {
    0xe1, 0xf9, 0x7a, 0x0d, 0x3e, 0x01, 0x8b, 0xe0,
    0xd6, 0x4f, 0xa3, 0x2c, 0x06, 0xde, 0x41, 0x39
  };
  unsigned char salt[14] = {
    0x0e, 0xc6, 0x75, 0xad, 0x49, 0x8a, 0xfe, 0xeb,
    0xb6, 0x96, 0x0b, 0x3a, 0xab, 0xe6
  };

  srtp_err(srtp_alloc(&srtp_sender,          /* srtp context                */
		      &rijndael_icm,	     /* cipher type                 */
		      16,		     /* cipher key length in octets */
		      &tmmhv2,		     /* authentication func type    */
		      94,		     /* auth key length in octets   */
		      4,		     /* auth tag length in octets   */
		      sec_serv_conf_and_auth /* security services flag      */
	     ));
  srtp_err(srtp_init_aes_128_prf(srtp_sender, test_key, salt));

  /* initialize data buffer */
  msg_len_octets = 28;
  hdr  = srtp_construct_test_packet(msg_len_octets
				    + srtp_get_trailer_length(srtp_sender));
  if (hdr == NULL)
    return err_status_alloc_fail;
  hdr2 = srtp_construct_test_packet(msg_len_octets
				    + srtp_get_trailer_length(srtp_sender)); 
  if (hdr2 == NULL) {
    free(hdr);
    return err_status_alloc_fail;
  }

  /* set message length */
  len = msg_len_octets;
  
  /* set pkt_end (for later checking for overruns) */
  pkt_end = (octet_t *)hdr +
    msg_len_octets + srtp_get_trailer_length(srtp_sender);

#if SRTP_VERBOSE
  printf("before protection:\npacket length: %d\n", len);
  srtp_print_packet(hdr, len);
#endif

  srtp_err(srtp_protect(srtp_sender, hdr, &len));

  /* check for overrun of the srtp_protect() function */
  for (i = 0; i < 4; i++)
    if (pkt_end[i] != 0xff) {
      fprintf(stdout, "overwrite in srtp_protect() function "
	      "(expected %x, found %x in trailing octet %d)\n",
	      0xff, ((octet_t *)hdr)[i], i);
      free(hdr);
      free(hdr2);
      return err_status_algo_fail;
    }  
  
#if SRTP_VERBOSE
  printf("after protection:\npacket length: %d\n", len);
  srtp_print_packet(hdr, len);
#endif
  
  srtp_err(srtp_alloc(&srtp_rcvr,            /* srtp context                */
		      &rijndael_icm,	     /* cipher type                 */
		      16,		     /* cipher key length in octets */
		      &tmmhv2,		     /* authentication func type    */
		      94,		     /* auth key length in octets   */
		      4,		     /* auth tag length in octets   */
		      sec_serv_conf_and_auth /* security services flag      */
	     ));
  srtp_err(srtp_init_aes_128_prf(srtp_rcvr, test_key, salt));
  
  srtp_err(srtp_unprotect(srtp_rcvr, hdr, &len));

#if SRTP_VERBOSE
  printf("after unprotection:\npacket length: %d\n", len);
  srtp_print_packet(hdr, msg_len_octets);
#endif
  fflush(stdout);

  for (i=0; i < msg_len_octets; i++)
    if (((octet_t *)hdr)[i] != ((octet_t *)hdr2)[i]) {
      fprintf(stdout, "mismatch at octet %d\n", i);
      status = err_status_algo_fail;
    }

  /* DAM - in the future, test for false negatives here */  

  return status;
}


