/*
 * ust-driver.c
 *
 * test driver for UST implementation
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */


#include <stdio.h>      /* for printf() */
#include <time.h>       /* for clock()  */

#include "ust.h"

extern cipher_type_t null_cipher;
extern cipher_type_t rijndael_icm;
extern auth_type_t   null_auth;
extern auth_type_t   tmmhv2;

#define MAX_MSG_OCTETS 1024
#define NUM_TRIALS     100000

double
ust_bits_per_second(ust_ctx_t *ctx, int msg_len_octets, int num_trials);

err_status_t
ust_test(ust_ctx_t *xfm_ctx, int msg_len_octets);

err_status_t
ust_auth_test();

int
main () {
  octet_t cipher_key[16] = {
    0x43, 0x3e, 0x69, 0x7c, 0xf8, 0x9d, 0x46, 0x82,
    0xe8, 0x98, 0x76, 0x58, 0xd5, 0x65, 0xb9, 0xbc
  };
 octet_t auth_key[94]   = {
   0x1f, 0x5a, 0x5f, 0xc0, 0x6b, 0x79, 0x4e, 0xbe,
   0x72, 0x38, 0x20, 0x53, 0x39, 0x5b, 0xbb, 0x9b,
   0xee, 0x22, 0x7c, 0x7f, 0x40, 0x3e, 0xb0, 0x5c,
   0xbd, 0x32, 0x3e, 0x80, 0x86, 0xa9, 0xad, 0x59,
   0x33, 0x19, 0x86, 0xcd, 0x25, 0x8b, 0xc6, 0x66,
   0xba, 0x45, 0x09, 0x6c, 0xdc, 0xaf, 0xa2, 0x49,
   0x81, 0x5f, 0xdb, 0x33, 0xa3, 0x43, 0xa7, 0x71,
   0xf8, 0xb6, 0xd1, 0xee, 0xb2, 0x23, 0xdf, 0xb4,
   0xb4, 0x0e, 0x5f, 0x10 
 };
  ust_ctx_t xfm_ctx;
  err_status_t stat;
  int len;
    
  stat = ust_init(&xfm_ctx,        /* ust context                       */  
		  6,               /* number of octets in the index     */
		  &rijndael_icm,   /* keystream generator specification */
		  cipher_key,      /* cipher key                        */
		  16,              /* number of octets in cipher key    */
		  &tmmhv2,         /* auth algorithm specification      */
		  auth_key,        /* auth key                          */
		  94,              /* number of octets in the auth key  */
		  4,               /* number of octets in the auth tag  */
		  128              /* length of replay window           */
		  );

  if (stat) {
    fprintf(stderr, "ust_xfm failed with error code %d\n", stat);
    exit(1);
  }

  printf("testing ust authentication function...");
  stat = ust_auth_test();
  if (stat) {
    fprintf(stderr, "ust_auth_test failed with error code %d\n", stat);
    exit(1);
  }
  printf("done\n");
  
  printf("testing ust_xfm and ust_inv_xfm...");
  stat = ust_test(&xfm_ctx, 64);  
  if (stat) {
    printf("ust_test failed with error code %d\n", stat);
    exit(1);
  } 
  printf("done\n");
  
  printf("timing ust_xfm()...\n");
  for (len=16; len < MAX_MSG_OCTETS; len *= 2) 
    printf("%d octets\t%f bits per second\n", 
	   len, ust_bits_per_second(&xfm_ctx, len, NUM_TRIALS) / 1e9 );
  
  return 0;
}

double
ust_bits_per_second(ust_ctx_t *xfm_ctx, int msg_len_octets, int num_trials) {
  clock_t timer;
  int i;
  xtd_seq_num_t idx;
  err_status_t stat;
  octet_t *msg; 


  /* allocate storage */
  msg = malloc(msg_len_octets);
  if (msg == NULL)
    return 0.0;  /* signal failure by returning zero */
  
  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  
  idx.roc = 0;
  idx.seq = 0;
  timer = clock();
  for(i=0; i < NUM_TRIALS; i++, index_advance(&idx, 1)) {
    stat = ust_xfm(xfm_ctx, idx, msg, msg_len_octets, 
		   msg, msg_len_octets, xfm_ctx->auth_tag);
    if (stat) {
      fprintf(stderr, "ust_xfm failed with error code %d\n", stat);
      exit(1);	
    }
  }
  timer = clock() - timer;

  free(msg);

  return (double) CLOCKS_PER_SEC * NUM_TRIALS * 8 * msg_len_octets / timer;
} 

/*
 * ust_test(ctx, len) creates a dummy message to which it applies 
 * a transform, then applies the inverse transform, verifying
 * that the data buffers before and after processing match, and that
 * the authentication works as well
 */

err_status_t
ust_test(ust_ctx_t *xfm_ctx, int msg_len_octets) {
  xtd_seq_num_t idx;
  err_status_t stat;
  octet_t *msg; 
  octet_t *tag;
  int i;

  /* allocate storage, set message to all nulls */
  msg = malloc(msg_len_octets + ust_tag_len(xfm_ctx));
  if (msg == NULL)
    return err_status_alloc_fail;
  /* the tag occupies the final portion of the newly allocated memory */
  tag = msg + msg_len_octets;
  for (i=0; i < msg_len_octets; i++)
    msg[i] = 0;
  
  /* check for proper decryption and false negatives */

  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  idx.roc = 0;
  idx.seq = 0;

  /* apply transform to msg */
  stat = ust_xfm(xfm_ctx, idx, msg, msg_len_octets, 
		 msg, msg_len_octets, tag);
  if (stat) {
    free(msg);
    return stat;
  }

  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  idx.roc = 0;
  idx.seq = 0;

  /* apply inverse transform to msg */
  stat = ust_inv_xfm(xfm_ctx, idx, msg, msg_len_octets, 
		     msg, msg_len_octets, tag);
  if (stat) {
    free(msg);
    return stat;
  }

  /* check that decrypted message is all-null */
  for (i=0; i < msg_len_octets; i++)
    if (msg[i] != 0x0000) {
      free(msg);
      return err_status_algo_fail;
    }

  /* now check for a false positive */

  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  idx.roc = 0;
  idx.seq = 0;

  /* apply transform to msg */
  stat = ust_xfm(xfm_ctx, idx, NULL, msg_len_octets, 
		 msg, msg_len_octets, tag);
  if (stat) {
    free(msg);
    return stat;
  }

  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  idx.roc = 0;
  idx.seq = 0;

  /* 
   * change message, then check for authentication failure
   */
  for (i=0; i < msg_len_octets; i++)
    msg[i] ^= 0xff;

  /* apply inverse transform to msg */
  stat = ust_inv_xfm(xfm_ctx, idx, NULL, msg_len_octets, 
		     msg, msg_len_octets, tag);
  if (stat != err_status_auth_fail) {
    free(msg);
    return err_status_algo_fail;
  }

  /* free memory and return success */
  free(msg);
  return err_status_ok;
} 


err_status_t
ust_auth_test() {
  uint16_t auth_key[47]   = {
    0x2337, 0x47e0, 0x1564, 0x671b, 0x6f80, 0xdcdd, 0xa6cc, 0x5ff1,
    0x3e5d, 0x88eb, 0x612e, 0x7c99, 0x02e8, 0xd8b2, 0x77a5, 0x09f9,
    0xf0bc, 0x9997, 0x1dc9, 0xd478, 0x396f, 0x9602, 0x8538, 0xaa7f,
    0x16a0, 0xa456, 0xe77e, 0x5262, 0xa1dc, 0x6b06, 0x5e67, 0xb2d5,
    0x74ee, 0x5045, 0x82c1, 0x310a, 0x28a5, 0xbb49, 0xf15a, 0x3834,
    0x59c8, 0xf3a, 0x36f7, 0x1b8c, 0x953d, 0xbf74, 0x2080 
  };
  uint16_t message[28] = {
    0x5741, 0x5220, 0x4953, 0x2050, 0x4541, 0x4345, 0x2c20, 0x4652,
    0x4545, 0x444f, 0x4d20, 0x4953, 0x2053, 0x4c41, 0x5645, 0x5259,
    0x2c20, 0x4947, 0x4e4f, 0x5241, 0x4e43, 0x4520, 0x4953, 0x2053,
    0x5452, 0x454e, 0x4754, 0x4800 
  };
  octet_t tag[4], ref_tag[4] = { 0x1D, 0x0F, 0x75, 0x36};
  xtd_seq_num_t idx;
  ust_ctx_t xfm_ctx;
  err_status_t stat;

  stat = ust_init(&xfm_ctx,        /* ust context                       */  
		  6,               /* number of octets in the index     */
		  &null_cipher,    /* keystream generator specification */
		  NULL,            /* cipher key                        */
		  0,               /* number of octets in cipher key    */
		  &tmmhv2,         /* auth algorithm specification      */
		  (octet_t *)auth_key,  /* auth key                       */
		  94,              /* number of octets in the auth key  */
		  4,               /* number of octets in the auth tag  */
		  128              /* length of replay window           */
		  );
   
  if (stat)
    return err_status_algo_fail;
      
  ust_xfm(&xfm_ctx, idx, NULL, 0, (octet_t *)message, 56, tag); 
  if (stat) 
    return err_status_algo_fail; 

  if (octet_string_is_eq(tag, ref_tag, 4))
    return err_status_algo_fail;
  
  ust_inv_xfm(&xfm_ctx, idx, NULL, 0, (octet_t *)message, 56, tag);
  if (stat)
    return err_status_algo_fail;
  
  return err_status_ok;
   
 }
